/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=M
  CPP_LANG=C+OPENMP
  CPP_PROCESSES_PER_NODE=saturno 1
*/

/* RECORD
Alberto Hernández Chillón
students of the University of Murcia, Spain
time 3489 msec
speed-up 11.44
*/ 

/*Nodes bounding has been included in the algorithm,
with an important reduction in the execution time.
No parallelism has been used, so the use of parallelism could
give aditional improvement.*/

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>

void writeall(int n,int p,int level,int valid,int value,int VOA,int *solution,int *cb)
{
  int i,j;
  printf("Level %d, value %d, optimum %d\n valid %d, solution:\n     ",level,value,VOA,valid);
  for(i=0;i<p;i++)printf("(%d,%d) ",solution[i]/n,solution[i]%n);
  printf("\nCheckerboard:\n");
  for(i=0;i<n;i++)
  {
    for(j=0;j<n;j++)printf("%3d ",cb[i*n+j]);
    printf("\n");
  }
  printf("\n");
}

void removepiece(int *piece,int level,int *solution,int n,int *cb)
{
	int i;

	cb[solution[level]]=0;
	for(i=1;i<(2*piece[0]-2);i+=2)
		cb[(solution[level]/n+piece[i])*n+solution[level]%n+piece[i+1]]=0;
}

void putpiece(int *piece,int level,int *solution,int n,int *cb)
{
	int i,ident=level+1;

	cb[solution[level]]=ident;
	for(i=1;i<(2*piece[0]-2);i+=2)
		cb[(solution[level]/n+piece[i])*n+solution[level]%n+piece[i+1]]=ident;
}

int feasiblepiece(int *piece,int level,int *solution,int n,int *cb)
{
	int i,valid=1,fil,col;

	if(cb[solution[level]]!=0)
		return 0;
	i=1;
	while(i<(2*piece[0]-2) && valid)
	{
		fil=solution[level]/n+piece[i];
		if(fil<0 || fil>(n-1))
			valid=0;
		else
		{
			col=solution[level]%n+piece[i+1];
			if(col<0 || col>(n-1))
				valid=0;
			else
				if(cb[fil*n+col]!=0)
					valid=0;
		}
		i+=2;
	}
	return valid;
}

void generate(int n,int p,int **pieces,int level,int *solution,int *value,int *cb,int *valid)
{
	if(solution[level]!=-1 && *valid)
	{
		removepiece(pieces[level],level,solution,n,cb);
		*value+=pieces[level][0];
	}
	solution[level]++;
	if(solution[level]==n*n)
		*valid=1;
	else
	{
		*valid=feasiblepiece(pieces[level],level,solution,n,cb);
		if(*valid)
		{
			putpiece(pieces[level],level,solution,n,cb);
			*value-=pieces[level][0];
		}
	}
}

int criterion(int p,int level,int valid)
{
	return(level!=p-1 && valid);
}

int morebrothers(int n,int nivel,int *solution)
{
	return(solution[nivel]<n*n);
}

void moveback(int n,int p,int **pieces,int level,int *solution,int *value,int *cb,int *valid)
{
	if(solution[level]!=n*n && *valid)
	{
		removepiece(pieces[level],level,solution,n,cb);
		*value+=pieces[level][0];
	}
	*valid=1;
	solution[level]=-1;
}

int realizarPoda(int n, int p, int **pieces, int level, int value, int VOA, int *solution, int *cb)
{
	int i, j, huecosBlanco = value, maxCasillasOcupadas = 0, colocada = 0;

	for(i = level; i < p; i++)
	{
		colocada = 0;
		for(j = 0; j < n*n; j++)
			if(cb[j] == i+1){colocada = 1; break;}
		if(!colocada)maxCasillasOcupadas = maxCasillasOcupadas + pieces[i][0];
	}

	/*printf("HUECOSENBLANCO = %d, MAXCASILLASOCUPADAS = %d\n", huecosBlanco, maxCasillasOcupadas);
	for(i=0; i < p; i++)printf("Solucion[%d]:%d\n", i, solution[i]);
	printf("RANDOM COSAS\n");
	printf("Dimension %d, piezas %d, Level %d, optimum %d, value: %d, solution:\n     ",n, p, level,VOA, value);
	for(i=0;i<p;i++)printf("(%d,%d) ",solution[i]/n,solution[i]%n);
	printf("\nCheckerboard:\n");
	for(i=0;i<n;i++){for(j=0;j<n;j++)printf("%3d ",cb[i*n+j]);printf("\n");}
	printf("\n");*/
	if((huecosBlanco - maxCasillasOcupadas) >= VOA){/*printf("Se poda\n");*/return 1;}
	//printf("No se poda\n");
	return 0;
}

int sec(int n,int p,int **pieces)
{
	int i, valid=1, value=n*n, VOA=n*n, level=0, *cb, *solution;

	cb=(int *) calloc(sizeof(int),n*n); //space for the checkerboard
	solution=(int *) malloc(sizeof(int)*p); //solution[i] is the position of the checkerboard where piece i is installed
	
	for(i=0;i<p;i++)solution[i]=-1; //indicates no decision on piece i is taken

	//printf("After Initialization\n"); writeall(n,p,level,valid,value,VOA,solution,cb);

	do{
		generate(n,p,pieces,level,solution,&value,cb,&valid);
		//printf("After Generate\n"); writeall(n,p,level,valid,value,VOA,solution,cb);

		if(valid && value<VOA)
		{
			VOA=value;
			//printf("After improving solution, the optimum is %d\n",VOA);
		}


		if(criterion(p,level,valid))
		{
			//printf("The criterion is valid\n");
			level++;
		}
		else
		{
			if(realizarPoda(n, p, pieces, level, value, VOA, solution, cb) && level !=-1)
			{
				moveback(n,p,pieces,level,solution,&value,cb,&valid);
				level--;
			}
      		while((level!=-1 && morebrothers(n,level,solution)==0))
			{
				//printf("After Morebrothers\n"); writeall(n,p,level,valid,value,VOA,solution,cb);
				moveback(n,p,pieces,level,solution,&value,cb,&valid);
				level--;
				//printf("After Moveback\n"); writeall(n,p,level,valid,value,VOA,solution,cb);
			}
		}
	} while(level!=-1);

	free(solution);
	free(cb);
	return VOA;
}
